"""Docker build utilities for Airbyte CDK."""

from __future__ import annotations

import json
import logging
import platform
import subprocess
import sys
from contextlib import ExitStack
from dataclasses import dataclass
from enum import Enum
from io import TextIOWrapper
from pathlib import Path

import click
import requests

from airbyte_cdk.models.connector_metadata import ConnectorLanguage, MetadataFile
from airbyte_cdk.test.entrypoint_wrapper import EntrypointOutput
from airbyte_cdk.utils.connector_paths import resolve_airbyte_repo_root


@dataclass(kw_only=True)
class ConnectorImageBuildError(Exception):
    """Custom exception for Docker build errors."""

    error_text: str
    build_args: list[str]

    def __str__(self) -> str:
        return "\n".join(
            [
                f"ConnectorImageBuildError: Could not build image.",
                f"Build args: {self.build_args}",
                f"Error text: {self.error_text}",
            ]
        )


logger = logging.getLogger(__name__)


class ArchEnum(str, Enum):
    """Enum for supported architectures."""

    ARM64 = "arm64"
    AMD64 = "amd64"


def _build_image(
    context_dir: Path,
    dockerfile: Path,
    metadata: MetadataFile,
    tag: str,
    arch: ArchEnum,
    build_args: dict[str, str | None] | None = None,
) -> str:
    """Build a Docker image for the specified architecture.

    Returns the tag of the built image.
    We use buildx to ensure we can build multi-platform images.

    Raises: ConnectorImageBuildError if the build fails.
    """
    docker_args: list[str] = [
        "docker",
        "buildx",
        "build",
        "--platform",
        f"linux/{arch.value}",
        "--file",
        str(dockerfile),
        "--label",
        f"io.airbyte.version={metadata.data.dockerImageTag}",
        "--label",
        f"io.airbyte.name={metadata.data.dockerRepository}",
    ]
    if build_args:
        for key, value in build_args.items():
            if value is not None:
                docker_args.extend(["--build-arg", f"{key}={value}"])
            else:
                docker_args.extend(["--build-arg", key])

    docker_args.extend(
        [
            "-t",
            tag,
            str(context_dir),
        ]
    )

    print(f"Building image: {tag} ({arch})")
    try:
        run_docker_command(
            docker_args,
            raise_if_errors=True,
            capture_stderr=True,
        )
    except subprocess.CalledProcessError as e:
        raise ConnectorImageBuildError(
            error_text=e.stderr,
            build_args=docker_args,
        ) from e

    return tag


def _tag_image(
    tag: str,
    new_tags: list[str] | str,
) -> None:
    """Build a Docker image for the specified architecture.

    Returns the tag of the built image.

    Raises:
        ConnectorImageBuildError: If the docker tag command fails.
    """
    if not isinstance(new_tags, list):
        new_tags = [new_tags]

    for new_tag in new_tags:
        print(f"Tagging image '{tag}' as: {new_tag}")
        docker_args = [
            "docker",
            "tag",
            tag,
            new_tag,
        ]
        try:
            run_docker_command(
                docker_args,
                raise_if_errors=True,
                capture_stderr=True,
            )
        except subprocess.CalledProcessError as e:
            raise ConnectorImageBuildError(
                error_text=e.stderr,
                build_args=docker_args,
            ) from e


def build_connector_image(
    connector_name: str,
    connector_directory: Path,
    *,
    metadata: MetadataFile,
    tag: str,
    no_verify: bool = False,
    dockerfile_override: Path | None = None,
) -> str:
    """Build a connector Docker image.

    This command builds a Docker image for a connector, using either
    the connector's Dockerfile or a base image specified in the metadata.
    The image is built for both AMD64 and ARM64 architectures.

    Args:
        connector_name: The name of the connector.
        connector_directory: The directory containing the connector code.
        metadata: The metadata of the connector.
        tag: The tag to apply to the built image.
        no_verify: If True, skip verification of the built image.

    Raises:
        ValueError: If the connector build options are not defined in metadata.yaml.
        ConnectorImageBuildError: If the image build or tag operation fails.
    """
    # Detect primary architecture based on the machine type.
    primary_arch: ArchEnum = (
        ArchEnum.ARM64
        if platform.machine().lower().startswith(("arm", "aarch"))
        else ArchEnum.AMD64
    )
    if not connector_name:
        raise ValueError("Connector name must be provided.")
    if not connector_directory:
        raise ValueError("Connector directory must be provided.")
    if not connector_directory.exists():
        raise ValueError(f"Connector directory does not exist: {connector_directory}")

    connector_kebab_name = connector_name
    connector_dockerfile_dir = connector_directory / "build" / "docker"

    if dockerfile_override:
        dockerfile_path = dockerfile_override
    else:
        dockerfile_path = connector_dockerfile_dir / "Dockerfile"
        dockerignore_path = connector_dockerfile_dir / "Dockerfile.dockerignore"
        try:
            dockerfile_text, dockerignore_text = get_dockerfile_templates(
                metadata=metadata,
                connector_directory=connector_directory,
            )
        except FileNotFoundError:
            # If the Dockerfile and .dockerignore are not found in the connector directory,
            # download the templates from the Airbyte repo. This is a fallback
            # in case the Airbyte repo not checked out locally.
            try:
                dockerfile_text, dockerignore_text = _download_dockerfile_defs(
                    connector_language=metadata.data.language,
                )
            except requests.HTTPError as e:
                raise ConnectorImageBuildError(
                    build_args=[],
                    error_text=(
                        "Could not locate local dockerfile templates and "
                        f"failed to download Dockerfile templates from github: {e}"
                    ),
                ) from e

        # ensure the directory exists
        connector_dockerfile_dir.mkdir(parents=True, exist_ok=True)
        dockerfile_path.write_text(dockerfile_text)
        dockerignore_path.write_text(dockerignore_text)

    extra_build_script: str = ""
    build_customization_path = connector_directory / "build_customization.py"
    if build_customization_path.exists():
        extra_build_script = str(build_customization_path)

    dockerfile_path.parent.mkdir(parents=True, exist_ok=True)
    if not metadata.data.connectorBuildOptions:
        raise ValueError(
            "Connector build options are not defined in metadata.yaml. "
            "Please check the connector's metadata file."
        )

    base_image = metadata.data.connectorBuildOptions.baseImage
    build_args: dict[str, str | None] = {
        "BASE_IMAGE": base_image,
        "CONNECTOR_NAME": connector_kebab_name,
        "EXTRA_BUILD_SCRIPT": extra_build_script,
    }

    base_tag = f"{metadata.data.dockerRepository}:{tag}"

    if metadata.data.language == ConnectorLanguage.JAVA:
        # This assumes that the repo root ('airbyte') is three levels above the
        # connector directory (airbyte/airbyte-integrations/connectors/source-foo).
        repo_root = connector_directory.parent.parent.parent
        # For Java connectors, we need to build the connector tar file first.
        subprocess.run(
            [
                "./gradlew",
                f":airbyte-integrations:connectors:{connector_name}:distTar",
            ],
            cwd=repo_root,
            text=True,
            check=True,
        )

    # Always build for AMD64, and optionally for ARM64 if needed locally.
    architectures = [ArchEnum.AMD64]
    if primary_arch == ArchEnum.ARM64:
        architectures += [ArchEnum.ARM64]

    built_images: list[str] = []
    for arch in architectures:
        docker_tag = f"{base_tag}-{arch.value}"
        docker_tag_parts = docker_tag.split("/")
        if len(docker_tag_parts) > 2:
            docker_tag = "/".join(docker_tag_parts[-1:])
        built_images.append(
            _build_image(
                context_dir=connector_directory,
                dockerfile=dockerfile_path,
                metadata=metadata,
                tag=docker_tag,
                arch=arch,
                build_args=build_args,
            )
        )

    _tag_image(
        tag=f"{base_tag}-{primary_arch.value}",
        new_tags=[base_tag],
    )
    if not no_verify:
        success, error_message = verify_connector_image(base_tag)
        if success:
            click.echo(f"Build and verification completed successfully: {base_tag}")
            return base_tag

        click.echo(
            f"Built image failed verification: {base_tag}\nError was:{error_message}", err=True
        )
        sys.exit(1)

    click.echo(f"Build completed successfully: {base_tag}")
    return base_tag


def _download_dockerfile_defs(
    connector_language: ConnectorLanguage,
) -> tuple[str, str]:
    """Download the Dockerfile and .dockerignore templates for the specified connector language.

    We use the requests library to download from the master branch hosted on GitHub.

    Args:
        connector_language: The language of the connector.

    Returns:
        A tuple containing the Dockerfile and .dockerignore templates as strings.

    Raises:
        ValueError: If the connector language is not supported.
        requests.HTTPError: If the download fails.
    """
    print("Downloading Dockerfile and .dockerignore templates from GitHub...")
    # Map ConnectorLanguage to template directory
    language_to_template_suffix = {
        ConnectorLanguage.PYTHON: "python-connector",
        ConnectorLanguage.JAVA: "java-connector",
        ConnectorLanguage.MANIFEST_ONLY: "manifest-only-connector",
    }

    if connector_language not in language_to_template_suffix:
        raise ValueError(f"Unsupported connector language: {connector_language}")

    template_suffix = language_to_template_suffix[connector_language]
    base_url = f"https://github.com/airbytehq/airbyte/raw/master/docker-images/"

    dockerfile_url = f"{base_url}/Dockerfile.{template_suffix}"
    dockerignore_url = f"{base_url}/Dockerfile.{template_suffix}.dockerignore"

    dockerfile_resp = requests.get(dockerfile_url)
    dockerfile_resp.raise_for_status()
    dockerfile_text = dockerfile_resp.text

    dockerignore_resp = requests.get(dockerignore_url)
    dockerignore_resp.raise_for_status()
    dockerignore_text = dockerignore_resp.text

    return dockerfile_text, dockerignore_text


def get_dockerfile_templates(
    metadata: MetadataFile,
    connector_directory: Path,
) -> tuple[str, str]:
    """Get the Dockerfile template for the connector.

    Args:
        metadata: The metadata of the connector.
        connector_name: The name of the connector.

    Raises:
        ValueError: If the connector language is not supported.
        FileNotFoundError: If the Dockerfile or .dockerignore is not found.

    Returns:
        A tuple containing the Dockerfile and .dockerignore templates as strings.
    """
    if metadata.data.language not in [
        ConnectorLanguage.PYTHON,
        ConnectorLanguage.MANIFEST_ONLY,
        ConnectorLanguage.JAVA,
    ]:
        raise ValueError(
            f"Unsupported connector language: {metadata.data.language}. "
            "Please check the connector's metadata file."
        )

    airbyte_repo_root = resolve_airbyte_repo_root(
        from_dir=connector_directory,
    )
    # airbyte_repo_root successfully resolved
    dockerfile_path = (
        airbyte_repo_root / "docker-images" / f"Dockerfile.{metadata.data.language.value}-connector"
    )
    dockerignore_path = (
        airbyte_repo_root
        / "docker-images"
        / f"Dockerfile.{metadata.data.language.value}-connector.dockerignore"
    )
    if not dockerfile_path.exists():
        raise FileNotFoundError(
            f"Dockerfile for {metadata.data.language.value} connector not found at {dockerfile_path}"
        )
    if not dockerignore_path.exists():
        raise FileNotFoundError(
            f".dockerignore for {metadata.data.language.value} connector not found at {dockerignore_path}"
        )

    return dockerfile_path.read_text(), dockerignore_path.read_text()


def run_docker_command(
    cmd: list[str],
    *,
    raise_if_errors: bool = True,
    capture_stdout: bool | Path = False,
    capture_stderr: bool | Path = False,
) -> subprocess.CompletedProcess[str]:
    """Run a Docker command as a subprocess.

    Note: When running Airbyte verbs such as `spec`, `discover`, `read`, etc.,
    use `run_docker_airbyte_command` instead to get an `EntrypointOutput` object as
    the return value and to better handle exceptions sent as messages.

    Args:
        cmd: The command to run as a list of strings.
        check: If True, raises an exception if the command fails. If False, the caller is
            responsible for checking the return code.
        capture_stdout: How to process stdout.
        capture_stderr: If True, captures stderr in memory and returns to the caller.
            If a Path is provided, the output is written to the specified file.

    For stdout and stderr process:
    - If False (the default), stdout is not captured.
    - If True, output is captured in memory and returned within the `CompletedProcess` object.
    - If a Path is provided, the output is written to the specified file. (Recommended for large syncs.)

    Raises:
        subprocess.CalledProcessError: If the command fails and check is True.
    """
    print(f"Running command: {' '.join(cmd)}")

    with ExitStack() as stack:
        # Shared context manager to handle file closing, if needed.
        stderr: TextIOWrapper | int | None
        stdout: TextIOWrapper | int | None

        # If capture_stderr or capture_stdout is a Path, we open the file in write mode.
        # If it's a boolean, we set it to either subprocess.PIPE or None.
        if isinstance(capture_stderr, Path):
            stderr = stack.enter_context(capture_stderr.open("w", encoding="utf-8"))
        elif isinstance(capture_stderr, bool):
            stderr = subprocess.PIPE if capture_stderr is True else None

        if isinstance(capture_stdout, Path):
            stdout = stack.enter_context(capture_stdout.open("w", encoding="utf-8"))
        elif isinstance(capture_stdout, bool):
            stdout = subprocess.PIPE if capture_stdout is True else None

        completed_process: subprocess.CompletedProcess[str] = subprocess.run(
            cmd,
            text=True,
            check=raise_if_errors,
            stderr=stderr,
            stdout=stdout,
        )
        return completed_process


def run_docker_airbyte_command(
    cmd: list[str],
    *,
    raise_if_errors: bool = False,
) -> EntrypointOutput:
    """Run an Airbyte command inside a Docker container.

    This wraps the `run_docker_command` function to process its results and
    return an `EntrypointOutput` object.

    Args:
        cmd: The command to run as a list of strings.
        raise_if_errors: If True, raises an exception if the command fails. If False, the caller is
            responsible for checking the for errors.

    Returns:
        The output of the command as an `EntrypointOutput` object.
    """
    process_result = run_docker_command(
        cmd,
        capture_stdout=True,
        capture_stderr=True,
        raise_if_errors=False,  # We want to handle failures ourselves.
    )
    result_output = EntrypointOutput(
        command=cmd,
        messages=process_result.stdout.splitlines(),
        uncaught_exception=(
            subprocess.CalledProcessError(
                cmd=cmd,
                returncode=process_result.returncode,
                output=process_result.stdout,
                stderr=process_result.stderr,
            )
            if process_result.returncode != 0
            else None
        ),
    )
    if raise_if_errors:
        # If check is True, we raise an exception if there are errors.
        # This will do nothing if there are no errors.
        result_output.raise_if_errors()

    return result_output


def verify_docker_installation() -> bool:
    """Verify Docker is installed and running."""
    try:
        run_docker_command(["docker", "--version"])
        return True
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False


def verify_connector_image(
    image_name: str,
) -> tuple[bool, str]:
    """Verify the built image by running the spec command.

    Args:
        image_name: The full image name with tag.

    Returns:
        True if the spec command succeeds, False otherwise.
    """
    logger.info(f"Verifying image {image_name} with 'spec' command...")

    try:
        result = run_docker_airbyte_command(
            ["docker", "run", "--rm", image_name, "spec"],
        )
        if result.errors:
            err_msg = result.get_formatted_error_message()
            logger.error(err_msg)
            return False, err_msg

        spec_messages = result.spec_messages
        if not spec_messages:
            err_msg = (
                "The container failed to produce valid output for the `spec` command.\nLog output:\n"
                + str(result.logs)
            )
            logger.error(err_msg)
            return False, err_msg

    except Exception as ex:
        err_msg = f"Unexpected error during image verification: {ex}"
        logger.error(err_msg)
        return False, err_msg

    return True, ""



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

from __future__ import annotations

from airbyte_cdk.sources.types import (
    Config,
    ConnectionDefinition,
    FieldPointer,
    Record,
    StreamSlice,
    StreamState,
)

# Note: This package originally contained class definitions for low-code CDK types, but we promoted them into the Python CDK.
# We've migrated connectors in the repository to reference the new location, but these assignments are used to retain backwards
# compatibility for sources created by OSS customers or on forks. This can be removed when we start bumping major versions.

FieldPointer = FieldPointer
Config = Config
ConnectionDefinition = ConnectionDefinition
StreamState = StreamState
Record = Record
StreamSlice = StreamSlice



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Pytest hooks for Airbyte CDK tests.

These hooks are used to customize the behavior of pytest during test discovery and execution.

To use these hooks within a connector, add the following lines to the connector's `conftest.py`
file, or to another file that is imported during test discovery:

```python
pytest_plugins = [
    "airbyte_cdk.test.standard_tests.pytest_hooks",
]
```
"""

from typing import Literal, cast

import pytest


@pytest.fixture
def connector_image_override(request: pytest.FixtureRequest) -> str | None:
    """Return the value of --connector-image, or None if not set."""
    return cast(str | None, request.config.getoption("--connector-image"))


@pytest.fixture
def read_from_streams(
    request: pytest.FixtureRequest,
) -> Literal["all", "none", "default"] | list[str]:
    """Specify if the test should read from streams.

    The input can be one of the following:
    - [Omitted] - Default to False, meaning no streams will be read.
    - `--read-from-streams`: Read from all suggested streams.
    - `--read-from-streams=true`: Read from all suggested streams.
    - `--read-from-streams=suggested`: Read from all suggested streams.
    - `--read-from-streams=default`: Read from all suggested streams.
    - `--read-from-streams=all`: Read from all streams.
    - `--read-from-streams=stream1,stream2`: Read from the specified streams only.
    - `--read-from-streams=false`: Do not read from any streams.
    - `--read-from-streams=none`: Do not read from any streams.
    """
    input_val: str = request.config.getoption(
        "--read-from-streams",
        default="default",  # type: ignore
    )  # type: ignore

    if isinstance(input_val, str):
        if input_val.lower() == "false":
            return "none"
        if input_val.lower() in ["true", "suggested", "default"]:
            # Default to 'default' (suggested) streams if the input is 'true', 'suggested', or
            # 'default'.
            # This is the default behavior if the option is not set.
            return "default"
        if input_val.lower() == "all":
            # This will sometimes fail if the account doesn't have permissions
            # to premium or restricted stream data.
            return "all"

        # If the input is a comma-separated list, split it into a list.
        # This will return a one-element list if the input is a single stream name.
        return input_val.split(",")

    # Else, probably a bool; return it as is.
    return input_val or "none"


@pytest.fixture
def read_scenarios(
    request: pytest.FixtureRequest,
) -> list[str] | Literal["all", "default"]:
    """Return the value of `--read-scenarios`.

    This argument is ignored if `--read-from-streams` is False or not set.

    The input can be one of the following:
    - [Omitted] - Default to 'config.json', meaning the default scenario will be read.
    - `--read-scenarios=all`: Read all scenarios.
    - `--read-scenarios=none`: Read no scenarios. (Overrides `--read-from-streams`, if set.)
    - `--read-scenarios=scenario1,scenario2`: Read the specified scenarios only.

    """
    input_val = cast(
        str,
        request.config.getoption(
            "--read-scenarios",
            default="default",  # type: ignore
        ),
    )

    if input_val.lower() == "default":
        # Default config scenario is always 'config.json'.
        return "default"

    if input_val.lower() == "none":
        # Default config scenario is always 'config.json'.
        return []

    return (
        [
            scenario_name.strip().lower().removesuffix(".json")
            for scenario_name in input_val.split(",")
        ]
        if input_val
        else []
    )


def pytest_addoption(parser: pytest.Parser) -> None:
    """Add --connector-image to pytest's CLI."""
    parser.addoption(
        "--connector-image",
        action="store",
        default=None,
        help="Use this pre-built connector Docker image instead of building one.",
    )
    parser.addoption(
        "--read-from-streams",
        action="store",
        default=None,
        help=read_from_streams.__doc__,
    )
    parser.addoption(
        "--read-scenarios",
        action="store",
        default="default",
        help=read_scenarios.__doc__,
    )


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    """A helper for pytest_generate_tests hook.

    If a test method (in a class subclassed from our base class)
    declares an argument 'scenario', this function retrieves the
    'scenarios' attribute from the test class and parametrizes that
    test with the values from 'scenarios'.

    ## Usage

    ```python
    from airbyte_cdk.test.standard_tests.connector_base import (
        generate_tests,
        ConnectorTestSuiteBase,
    )

    def pytest_generate_tests(metafunc):
        generate_tests(metafunc)

    class TestMyConnector(ConnectorTestSuiteBase):
        ...

    ```
    """
    # Check if the test function requires an 'scenario' argument
    if "scenario" in metafunc.fixturenames:
        # Retrieve the test class
        test_class = metafunc.cls
        if test_class is None:
            return

        # Check that the class is compatible with our test suite
        scenarios_attr = getattr(test_class, "get_scenarios", None)
        if scenarios_attr is None:
            raise ValueError(
                f"Test class {test_class} does not have a 'scenarios' attribute. "
                "Please define the 'scenarios' attribute in the test class."
            )

        # Get the scenarios defined or discovered in the test class
        scenarios = test_class.get_scenarios()

        # Create pytest.param objects with special marks as needed
        parametrized_scenarios = [
            pytest.param(
                scenario,
                marks=[pytest.mark.requires_creds] if scenario.requires_creds else [],
            )
            for scenario in scenarios
        ]

        # Parametrize the 'scenario' argument with the scenarios
        metafunc.parametrize(
            "scenario",
            parametrized_scenarios,
            ids=[str(scenario) for scenario in scenarios],
        )



#
# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
#

from dataclasses import dataclass

from airbyte_cdk.sources.declarative.types import Record

from .default_file_uploader import DefaultFileUploader
from .file_uploader import FileUploader


@dataclass
class ConnectorBuilderFileUploader(FileUploader):
    """
    Connector builder file uploader
    Acts as a decorator or wrapper around a FileUploader instance, copying the attributes from record.file_reference into the record.data.
    """

    file_uploader: DefaultFileUploader

    def upload(self, record: Record) -> None:
        self.file_uploader.upload(record=record)
        record.data["source_file_relative_path"] = record.file_reference.source_file_relative_path  # type: ignore



# Copyright (c) 2025 Airbyte, Inc., all rights reserved.

import json
import logging
import pkgutil
from copy import deepcopy
from dataclasses import dataclass, field
from queue import Queue
from types import ModuleType
from typing import (
    Any,
    ClassVar,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Set,
)

import orjson
import yaml
from airbyte_protocol_dataclasses.models import AirbyteStreamStatus, Level, StreamDescriptor
from jsonschema.exceptions import ValidationError
from jsonschema.validators import validate

from airbyte_cdk.config_observation import create_connector_config_control_message
from airbyte_cdk.connector_builder.models import (
    LogMessage as ConnectorBuilderLogMessage,
)
from airbyte_cdk.manifest_migrations.migration_handler import (
    ManifestMigrationHandler,
)
from airbyte_cdk.models import (
    AirbyteCatalog,
    AirbyteConnectionStatus,
    AirbyteMessage,
    AirbyteStateMessage,
    ConfiguredAirbyteCatalog,
    ConnectorSpecification,
    FailureType,
    Status,
)
from airbyte_cdk.models.airbyte_protocol_serializers import AirbyteMessageSerializer
from airbyte_cdk.sources import Source
from airbyte_cdk.sources.concurrent_source.concurrent_source import ConcurrentSource
from airbyte_cdk.sources.connector_state_manager import ConnectorStateManager
from airbyte_cdk.sources.declarative.checks import COMPONENTS_CHECKER_TYPE_MAPPING
from airbyte_cdk.sources.declarative.checks.connection_checker import ConnectionChecker
from airbyte_cdk.sources.declarative.concurrency_level import ConcurrencyLevel
from airbyte_cdk.sources.declarative.interpolation import InterpolatedBoolean
from airbyte_cdk.sources.declarative.models.declarative_component_schema import (
    ConcurrencyLevel as ConcurrencyLevelModel,
)
from airbyte_cdk.sources.declarative.models.declarative_component_schema import (
    DeclarativeStream as DeclarativeStreamModel,
)
from airbyte_cdk.sources.declarative.models.declarative_component_schema import (
    Spec as SpecModel,
)
from airbyte_cdk.sources.declarative.models.declarative_component_schema import (
    StateDelegatingStream as StateDelegatingStreamModel,
)
from airbyte_cdk.sources.declarative.parsers.custom_code_compiler import (
    get_registered_components_module,
)
from airbyte_cdk.sources.declarative.parsers.manifest_component_transformer import (
    ManifestComponentTransformer,
)
from airbyte_cdk.sources.declarative.parsers.manifest_normalizer import (
    ManifestNormalizer,
)
from airbyte_cdk.sources.declarative.parsers.manifest_reference_resolver import (
    ManifestReferenceResolver,
)
from airbyte_cdk.sources.declarative.parsers.model_to_component_factory import (
    ModelToComponentFactory,
)
from airbyte_cdk.sources.declarative.resolvers import COMPONENTS_RESOLVER_TYPE_MAPPING
from airbyte_cdk.sources.declarative.spec.spec import Spec
from airbyte_cdk.sources.declarative.types import Config, ConnectionDefinition
from airbyte_cdk.sources.message.concurrent_repository import ConcurrentMessageRepository
from airbyte_cdk.sources.message.repository import InMemoryMessageRepository
from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream
from airbyte_cdk.sources.streams.concurrent.partitions.types import QueueItem
from airbyte_cdk.sources.utils.slice_logger import (
    AlwaysLogSliceLogger,
    DebugSliceLogger,
    SliceLogger,
)
from airbyte_cdk.utils.stream_status_utils import as_airbyte_message
from airbyte_cdk.utils.traced_exception import AirbyteTracedException


@dataclass
class TestLimits:
    __test__: ClassVar[bool] = False  # Tell Pytest this is not a Pytest class, despite its name

    DEFAULT_MAX_PAGES_PER_SLICE: ClassVar[int] = 5
    DEFAULT_MAX_SLICES: ClassVar[int] = 5
    DEFAULT_MAX_RECORDS: ClassVar[int] = 100
    DEFAULT_MAX_STREAMS: ClassVar[int] = 100

    max_records: int = field(default=DEFAULT_MAX_RECORDS)
    max_pages_per_slice: int = field(default=DEFAULT_MAX_PAGES_PER_SLICE)
    max_slices: int = field(default=DEFAULT_MAX_SLICES)
    max_streams: int = field(default=DEFAULT_MAX_STREAMS)


def _get_declarative_component_schema() -> Dict[str, Any]:
    try:
        raw_component_schema = pkgutil.get_data(
            "airbyte_cdk", "sources/declarative/declarative_component_schema.yaml"
        )
        if raw_component_schema is not None:
            declarative_component_schema = yaml.load(raw_component_schema, Loader=yaml.SafeLoader)
            return declarative_component_schema  # type: ignore
        else:
            raise RuntimeError(
                "Failed to read manifest component json schema required for deduplication"
            )
    except FileNotFoundError as e:
        raise FileNotFoundError(
            f"Failed to read manifest component json schema required for deduplication: {e}"
        )


class ConcurrentDeclarativeSource(Source):
    # By default, we defer to a value of 2. A value lower than could cause a PartitionEnqueuer to be stuck in a state of deadlock
    # because it has hit the limit of futures but not partition reader is consuming them.
    _LOWEST_SAFE_CONCURRENCY_LEVEL = 2

    def __init__(
        self,
        catalog: Optional[ConfiguredAirbyteCatalog] = None,
        config: Optional[Mapping[str, Any]] = None,
        state: Optional[List[AirbyteStateMessage]] = None,
        *,
        source_config: ConnectionDefinition,
        debug: bool = False,
        emit_connector_builder_messages: bool = False,
        migrate_manifest: bool = False,
        normalize_manifest: bool = False,
        limits: Optional[TestLimits] = None,
        config_path: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self.logger = logging.getLogger(f"airbyte.{self.name}")

        self._limits = limits

        # todo: We could remove state from initialization. Now that streams are grouped during the read(), a source
        #  no longer needs to store the original incoming state. But maybe there's an edge case?
        self._connector_state_manager = ConnectorStateManager(state=state)  # type: ignore  # state is always in the form of List[AirbyteStateMessage]. The ConnectorStateManager should use generics, but this can be done later

        # We set a maxsize to for the main thread to process record items when the queue size grows. This assumes that there are less
        # threads generating partitions that than are max number of workers. If it weren't the case, we could have threads only generating
        # partitions which would fill the queue. This number is arbitrarily set to 10_000 but will probably need to be changed given more
        # information and might even need to be configurable depending on the source
        queue: Queue[QueueItem] = Queue(maxsize=10_000)
        message_repository = InMemoryMessageRepository(
            Level.DEBUG if emit_connector_builder_messages else Level.INFO
        )

        # To reduce the complexity of the concurrent framework, we are not enabling RFR with synthetic
        # cursors. We do this by no longer automatically instantiating RFR cursors when converting
        # the declarative models into runtime components. Concurrent sources will continue to checkpoint
        # incremental streams running in full refresh.
        component_factory = ModelToComponentFactory(
            emit_connector_builder_messages=emit_connector_builder_messages,
            message_repository=ConcurrentMessageRepository(queue, message_repository),
            configured_catalog=catalog,
            connector_state_manager=self._connector_state_manager,
            max_concurrent_async_job_count=source_config.get("max_concurrent_async_job_count"),
            limit_pages_fetched_per_slice=limits.max_pages_per_slice if limits else None,
            limit_slices_fetched=limits.max_slices if limits else None,
            disable_retries=True if limits else False,
            disable_cache=True if limits else False,
        )

        self._should_normalize = normalize_manifest
        self._should_migrate = migrate_manifest
        self._declarative_component_schema = _get_declarative_component_schema()
        # If custom components are needed, locate and/or register them.
        self.components_module: ModuleType | None = get_registered_components_module(config=config)
        # set additional attributes
        self._debug = debug
        self._emit_connector_builder_messages = emit_connector_builder_messages
        self._constructor = (
            component_factory
            if component_factory
            else ModelToComponentFactory(
                emit_connector_builder_messages=emit_connector_builder_messages,
                max_concurrent_async_job_count=source_config.get("max_concurrent_async_job_count"),
            )
        )

        self._message_repository = self._constructor.get_message_repository()
        self._slice_logger: SliceLogger = (
            AlwaysLogSliceLogger() if emit_connector_builder_messages else DebugSliceLogger()
        )

        # resolve all components in the manifest
        self._source_config = self._pre_process_manifest(dict(source_config))
        # validate resolved manifest against the declarative component schema
        self._validate_source()
        # apply additional post-processing to the manifest
        self._post_process_manifest()

        spec: Optional[Mapping[str, Any]] = self._source_config.get("spec")
        self._spec_component: Optional[Spec] = (
            self._constructor.create_component(SpecModel, spec, dict()) if spec else None
        )
        self._config = self._migrate_and_transform_config(config_path, config) or {}

        concurrency_level_from_manifest = self._source_config.get("concurrency_level")
        if concurrency_level_from_manifest:
            concurrency_level_component = self._constructor.create_component(
                model_type=ConcurrencyLevelModel,
                component_definition=concurrency_level_from_manifest,
                config=config or {},
            )
            if not isinstance(concurrency_level_component, ConcurrencyLevel):
                raise ValueError(
                    f"Expected to generate a ConcurrencyLevel component, but received {concurrency_level_component.__class__}"
                )

            concurrency_level = concurrency_level_component.get_concurrency_level()
            initial_number_of_partitions_to_generate = max(
                concurrency_level // 2, 1
            )  # Partition_generation iterates using range based on this value. If this is floored to zero we end up in a dead lock during start up
        else:
            concurrency_level = self._LOWEST_SAFE_CONCURRENCY_LEVEL
            initial_number_of_partitions_to_generate = self._LOWEST_SAFE_CONCURRENCY_LEVEL // 2

        self._concurrent_source = ConcurrentSource.create(
            num_workers=concurrency_level,
            initial_number_of_partitions_to_generate=initial_number_of_partitions_to_generate,
            logger=self.logger,
            slice_logger=self._slice_logger,
            queue=queue,
            message_repository=self._message_repository,
        )

    def _pre_process_manifest(self, manifest: Dict[str, Any]) -> Dict[str, Any]:
        """
        Preprocesses the provided manifest dictionary by resolving any manifest references.

        This method modifies the input manifest in place, resolving references using the
        ManifestReferenceResolver to ensure all references within the manifest are properly handled.

        Args:
            manifest (Dict[str, Any]): The manifest dictionary to preprocess and resolve references in.

        Returns:
            None
        """
        # For ease of use we don't require the type to be specified at the top level manifest, but it should be included during processing
        manifest = self._fix_source_type(manifest)
        # Resolve references in the manifest
        resolved_manifest = ManifestReferenceResolver().preprocess_manifest(manifest)
        # Propagate types and parameters throughout the manifest
        propagated_manifest = ManifestComponentTransformer().propagate_types_and_parameters(
            "", resolved_manifest, {}
        )

        return propagated_manifest

    def _fix_source_type(self, manifest: Dict[str, Any]) -> Dict[str, Any]:
        """
        Fix the source type in the manifest. This is necessary because the source type is not always set in the manifest.
        """
        if "type" not in manifest:
            manifest["type"] = "DeclarativeSource"

        return manifest

    def _post_process_manifest(self) -> None:
        """
        Post-processes the manifest after validation.
        This method is responsible for any additional modifications or transformations needed
        after the manifest has been validated and before it is used in the source.
        """
        # apply manifest migration, if required
        self._migrate_manifest()
        # apply manifest normalization, if required
        self._normalize_manifest()

    def _migrate_manifest(self) -> None:
        """
        This method is used to migrate the manifest. It should be called after the manifest has been validated.
        The migration is done in place, so the original manifest is modified.

        The original manifest is returned if any error occurs during migration.
        """
        if self._should_migrate:
            manifest_migrator = ManifestMigrationHandler(self._source_config)
            self._source_config = manifest_migrator.apply_migrations()
            # validate migrated manifest against the declarative component schema
            self._validate_source()

    def _normalize_manifest(self) -> None:
        """
        This method is used to normalize the manifest. It should be called after the manifest has been validated.

        Connector Builder UI rendering requires the manifest to be in a specific format.
         - references have been resolved
         - the commonly used definitions are extracted to the `definitions.linked.*`
        """
        if self._should_normalize:
            normalizer = ManifestNormalizer(self._source_config, self._declarative_component_schema)
            self._source_config = normalizer.normalize()

    def _validate_source(self) -> None:
        """
        Validates the connector manifest against the declarative component schema
        """

        try:
            validate(self._source_config, self._declarative_component_schema)
        except ValidationError as e:
            raise ValidationError(
                "Validation against json schema defined in declarative_component_schema.yaml schema failed"
            ) from e

    def _migrate_and_transform_config(
        self,
        config_path: Optional[str],
        config: Optional[Config],
    ) -> Optional[Config]:
        if not config:
            return None
        if not self._spec_component:
            return config
        mutable_config = dict(config)
        self._spec_component.migrate_config(mutable_config)
        if mutable_config != config:
            if config_path:
                with open(config_path, "w") as f:
                    json.dump(mutable_config, f)
            control_message = create_connector_config_control_message(mutable_config)
            print(orjson.dumps(AirbyteMessageSerializer.dump(control_message)).decode())
        self._spec_component.transform_config(mutable_config)
        return mutable_config

    def configure(self, config: Mapping[str, Any], temp_dir: str) -> Mapping[str, Any]:
        config = self._config or config
        return super().configure(config, temp_dir)

    @property
    def resolved_manifest(self) -> Mapping[str, Any]:
        """
        Returns the resolved manifest configuration for the source.

        This property provides access to the internal source configuration as a mapping,
        which contains all settings and parameters required to define the source's behavior.

        Returns:
            Mapping[str, Any]: The resolved source configuration manifest.
        """
        return self._source_config

    def deprecation_warnings(self) -> List[ConnectorBuilderLogMessage]:
        return self._constructor.get_model_deprecations()

    def read(
        self,
        logger: logging.Logger,
        config: Mapping[str, Any],
        catalog: ConfiguredAirbyteCatalog,
        state: Optional[List[AirbyteStateMessage]] = None,
    ) -> Iterator[AirbyteMessage]:
        selected_concurrent_streams = self._select_streams(
            streams=self.streams(config=self._config),  # type: ignore  # We are migrating away from the DeclarativeStream implementation and streams() only returns the concurrent-compatible AbstractStream. To preserve compatibility, we retain the existing method interface
            configured_catalog=catalog,
        )

        # It would appear that passing in an empty set of streams causes an infinite loop in ConcurrentReadProcessor.
        # This is also evident in concurrent_source_adapter.py so I'll leave this out of scope to fix for now
        if len(selected_concurrent_streams) > 0:
            yield from self._concurrent_source.read(selected_concurrent_streams)

    def discover(self, logger: logging.Logger, config: Mapping[str, Any]) -> AirbyteCatalog:
        return AirbyteCatalog(
            streams=[stream.as_airbyte_stream() for stream in self.streams(config=self._config)]
        )

    # todo: add PR comment about whether we can change the signature to List[AbstractStream]
    def streams(self, config: Mapping[str, Any]) -> List[AbstractStream]:  # type: ignore  # we are migrating away from the AbstractSource and are expecting that this will only be called by ConcurrentDeclarativeSource or the Connector Builder
        """
        The `streams` method is used as part of the AbstractSource in the following cases:
        * ConcurrentDeclarativeSource.check -> ManifestDeclarativeSource.check -> AbstractSource.check -> DeclarativeSource.check_connection -> CheckStream.check_connection -> streams
        * ConcurrentDeclarativeSource.read -> AbstractSource.read -> streams (note that we filter for a specific catalog which excludes concurrent streams so not all streams actually read from all the streams returned by `streams`)
        Note that `super.streams(config)` is also called when splitting the streams between concurrent or not in `_group_streams`.

        In both case, we will assume that calling the DeclarativeStream is perfectly fine as the result for these is the same regardless of if it is a DeclarativeStream or a DefaultStream (concurrent). This should simply be removed once we have moved away from the mentioned code paths above.
        """

        if self._spec_component:
            self._spec_component.validate_config(self._config)

        stream_configs = self._stream_configs(self._source_config) + self.dynamic_streams

        api_budget_model = self._source_config.get("api_budget")
        if api_budget_model:
            self._constructor.set_api_budget(api_budget_model, self._config)

        source_streams = [
            self._constructor.create_component(
                (
                    StateDelegatingStreamModel
                    if stream_config.get("type") == StateDelegatingStreamModel.__name__
                    else DeclarativeStreamModel
                ),
                stream_config,
                self._config,
                emit_connector_builder_messages=self._emit_connector_builder_messages,
            )
            for stream_config in self._initialize_cache_for_parent_streams(deepcopy(stream_configs))
        ]
        return source_streams

    @staticmethod
    def _initialize_cache_for_parent_streams(
        stream_configs: List[Dict[str, Any]],
    ) -> List[Dict[str, Any]]:
        """Enable caching for parent streams unless explicitly disabled.

        Caching is enabled by default for parent streams to optimize performance when the same
        parent data is needed by multiple child streams. However, explicit `use_cache: false`
        settings are respected for streams that cannot use caching (e.g., scroll-based pagination
        APIs where caching causes duplicate records).
        """
        parent_streams = set()

        def _set_cache_if_not_disabled(requester: Dict[str, Any]) -> None:
            """Set use_cache to True only if not explicitly disabled."""
            if requester.get("use_cache") is not False:
                requester["use_cache"] = True

        def update_with_cache_parent_configs(
            parent_configs: list[dict[str, Any]],
        ) -> None:
            for parent_config in parent_configs:
                parent_streams.add(parent_config["stream"]["name"])
                if parent_config["stream"]["type"] == "StateDelegatingStream":
                    _set_cache_if_not_disabled(
                        parent_config["stream"]["full_refresh_stream"]["retriever"]["requester"]
                    )
                    _set_cache_if_not_disabled(
                        parent_config["stream"]["incremental_stream"]["retriever"]["requester"]
                    )
                else:
                    _set_cache_if_not_disabled(parent_config["stream"]["retriever"]["requester"])

        for stream_config in stream_configs:
            if stream_config.get("incremental_sync", {}).get("parent_stream"):
                parent_streams.add(stream_config["incremental_sync"]["parent_stream"]["name"])
                _set_cache_if_not_disabled(
                    stream_config["incremental_sync"]["parent_stream"]["retriever"]["requester"]
                )

            elif stream_config.get("retriever", {}).get("partition_router", {}):
                partition_router = stream_config["retriever"]["partition_router"]

                if isinstance(partition_router, dict) and partition_router.get(
                    "parent_stream_configs"
                ):
                    update_with_cache_parent_configs(partition_router["parent_stream_configs"])
                elif isinstance(partition_router, list):
                    for router in partition_router:
                        if router.get("parent_stream_configs"):
                            update_with_cache_parent_configs(router["parent_stream_configs"])

        for stream_config in stream_configs:
            if stream_config["name"] in parent_streams:
                if stream_config["type"] == "StateDelegatingStream":
                    _set_cache_if_not_disabled(
                        stream_config["full_refresh_stream"]["retriever"]["requester"]
                    )
                    _set_cache_if_not_disabled(
                        stream_config["incremental_stream"]["retriever"]["requester"]
                    )
                else:
                    _set_cache_if_not_disabled(stream_config["retriever"]["requester"])
        return stream_configs

    def spec(self, logger: logging.Logger) -> ConnectorSpecification:
        """
        Returns the connector specification (spec) as defined in the Airbyte Protocol. The spec is an object describing the possible
        configurations (e.g: username and password) which can be configured when running this connector. For low-code connectors, this
        will first attempt to load the spec from the manifest's spec block, otherwise it will load it from "spec.yaml" or "spec.json"
        in the project root.
        """
        return (
            self._spec_component.generate_spec() if self._spec_component else super().spec(logger)
        )

    def check(self, logger: logging.Logger, config: Mapping[str, Any]) -> AirbyteConnectionStatus:
        check = self._source_config.get("check")
        if not check:
            raise ValueError(f"Missing 'check' component definition within the manifest.")

        if "type" not in check:
            check["type"] = "CheckStream"
        connection_checker = self._constructor.create_component(
            COMPONENTS_CHECKER_TYPE_MAPPING[check["type"]],
            check,
            dict(),
            emit_connector_builder_messages=self._emit_connector_builder_messages,
        )
        if not isinstance(connection_checker, ConnectionChecker):
            raise ValueError(
                f"Expected to generate a ConnectionChecker component, but received {connection_checker.__class__}"
            )

        check_succeeded, error = connection_checker.check_connection(self, logger, self._config)
        if not check_succeeded:
            return AirbyteConnectionStatus(status=Status.FAILED, message=repr(error))
        return AirbyteConnectionStatus(status=Status.SUCCEEDED)

    @property
    def dynamic_streams(self) -> List[Dict[str, Any]]:
        return self._dynamic_stream_configs(
            manifest=self._source_config,
            with_dynamic_stream_name=True,
        )

    def _stream_configs(self, manifest: Mapping[str, Any]) -> List[Dict[str, Any]]:
        # This has a warning flag for static, but after we finish part 4 we'll replace manifest with self._source_config
        stream_configs = []
        for current_stream_config in manifest.get("streams", []):
            if (
                "type" in current_stream_config
                and current_stream_config["type"] == "ConditionalStreams"
            ):
                interpolated_boolean = InterpolatedBoolean(
                    condition=current_stream_config.get("condition"),
                    parameters={},
                )

                if interpolated_boolean.eval(config=self._config):
                    stream_configs.extend(current_stream_config.get("streams", []))
            else:
                if "type" not in current_stream_config:
                    current_stream_config["type"] = "DeclarativeStream"
                stream_configs.append(current_stream_config)
        return stream_configs

    def _dynamic_stream_configs(
        self,
        manifest: Mapping[str, Any],
        with_dynamic_stream_name: Optional[bool] = None,
    ) -> List[Dict[str, Any]]:
        dynamic_stream_definitions: List[Dict[str, Any]] = manifest.get("dynamic_streams", [])
        dynamic_stream_configs: List[Dict[str, Any]] = []
        seen_dynamic_streams: Set[str] = set()

        for dynamic_definition_index, dynamic_definition in enumerate(dynamic_stream_definitions):
            components_resolver_config = dynamic_definition["components_resolver"]

            if not components_resolver_config:
                raise ValueError(
                    f"Missing 'components_resolver' in dynamic definition: {dynamic_definition}"
                )

            resolver_type = components_resolver_config.get("type")
            if not resolver_type:
                raise ValueError(
                    f"Missing 'type' in components resolver configuration: {components_resolver_config}"
                )

            if resolver_type not in COMPONENTS_RESOLVER_TYPE_MAPPING:
                raise ValueError(
                    f"Invalid components resolver type '{resolver_type}'. "
                    f"Expected one of {list(COMPONENTS_RESOLVER_TYPE_MAPPING.keys())}."
                )

            if "retriever" in components_resolver_config:
                components_resolver_config["retriever"]["requester"]["use_cache"] = True

            # Create a resolver for dynamic components based on type
            if resolver_type == "HttpComponentsResolver":
                components_resolver = self._constructor.create_component(
                    model_type=COMPONENTS_RESOLVER_TYPE_MAPPING[resolver_type],
                    component_definition=components_resolver_config,
                    config=self._config,
                    stream_name=dynamic_definition.get("name"),
                )
            else:
                components_resolver = self._constructor.create_component(
                    model_type=COMPONENTS_RESOLVER_TYPE_MAPPING[resolver_type],
                    component_definition=components_resolver_config,
                    config=self._config,
                )

            stream_template_config = dynamic_definition["stream_template"]

            for dynamic_stream in components_resolver.resolve_components(
                stream_template_config=stream_template_config
            ):
                # Get the use_parent_parameters configuration from the dynamic definition
                # Default to True for backward compatibility, since connectors were already using it by default when this param was added
                use_parent_parameters = dynamic_definition.get("use_parent_parameters", True)

                dynamic_stream = {
                    **ManifestComponentTransformer().propagate_types_and_parameters(
                        "", dynamic_stream, {}, use_parent_parameters=use_parent_parameters
                    )
                }

                if "type" not in dynamic_stream:
                    dynamic_stream["type"] = "DeclarativeStream"

                # Ensure that each stream is created with a unique name
                name = dynamic_stream.get("name")

                if with_dynamic_stream_name:
                    dynamic_stream["dynamic_stream_name"] = dynamic_definition.get(
                        "name", f"dynamic_stream_{dynamic_definition_index}"
                    )

                if not isinstance(name, str):
                    raise ValueError(
                        f"Expected stream name {name} to be a string, got {type(name)}."
                    )

                if name in seen_dynamic_streams:
                    error_message = f"Dynamic streams list contains a duplicate name: {name}. Please contact Airbyte Support."
                    failure_type = FailureType.system_error

                    if resolver_type == "ConfigComponentsResolver":
                        error_message = f"Dynamic streams list contains a duplicate name: {name}. Please check your configuration."
                        failure_type = FailureType.config_error

                    raise AirbyteTracedException(
                        message=error_message,
                        internal_message=error_message,
                        failure_type=failure_type,
                    )

                seen_dynamic_streams.add(name)
                dynamic_stream_configs.append(dynamic_stream)

        return dynamic_stream_configs

    def _select_streams(
        self, streams: List[AbstractStream], configured_catalog: ConfiguredAirbyteCatalog
    ) -> List[AbstractStream]:
        stream_name_to_instance: Mapping[str, AbstractStream] = {s.name: s for s in streams}
        abstract_streams: List[AbstractStream] = []
        for configured_stream in configured_catalog.streams:
            stream_instance = stream_name_to_instance.get(configured_stream.stream.name)
            if stream_instance:
                abstract_streams.append(stream_instance)
            else:
                # Previous behavior in the legacy synchronous CDK was to also raise an error TRACE message if
                # the source was configured with raise_exception_on_missing_stream=True. This was used on very
                # few sources like facebook-marketing and google-ads. We decided not to port this feature over,
                # but we can do so if we feel it necessary. With the current behavior,we should still result
                # in a partial failure since missing streams will be marked as INCOMPLETE.
                self._message_repository.emit_message(
                    as_airbyte_message(configured_stream.stream, AirbyteStreamStatus.INCOMPLETE)
                )
        return abstract_streams



import os
from typing import Any, Dict

from fastapi import APIRouter

from ..api_models.capabilities import CapabilitiesResponse

router = APIRouter(
    prefix="/capabilities",
    tags=["capabilities"],
)


@router.get("/", operation_id="getCapabilities")
def get_capabilities() -> CapabilitiesResponse:
    """
    Get the capabilities available for the manifest server.

    Returns:
        Dict containing the service capabilities including custom code execution support.
    """
    # Read the same environment variable as the connector builder server
    enable_unsafe_code = os.getenv("AIRBYTE_ENABLE_UNSAFE_CODE", "false").lower() == "true"

    return CapabilitiesResponse(custom_code_execution=enable_unsafe_code)



import logging
from typing import Optional

import ddtrace

logger = logging.getLogger(__name__)


def apply_trace_tags_from_context(
    workspace_id: Optional[str] = None,
    project_id: Optional[str] = None,
) -> None:
    """Apply trace tags from context to the current span."""
    if not workspace_id and not project_id:
        return

    # Log the trace IDs for observability
    log_parts = []
    if workspace_id:
        log_parts.append(f"workspace_id={workspace_id}")
    if project_id:
        log_parts.append(f"project_id={project_id}")

    if log_parts:
        logger.info(f"Processing request with trace tags: {', '.join(log_parts)}")

    try:
        span = ddtrace.tracer.current_span()
        if span:
            if workspace_id:
                span.set_tag("workspace_id", workspace_id)
            if project_id:
                span.set_tag("project_id", project_id)
    except Exception:
        # Silently ignore any ddtrace-related errors (e.g. if ddtrace.auto wasn't run)
        pass



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

# The goal of this variable is to make an implicit dependency explicit. As part of of the Concurrent CDK work, we are facing a situation
# where the connection pool size is too small to serve all the threads (see https://github.com/airbytehq/airbyte/issues/32072). In
# order to fix that, we will increase the requests library pool_maxsize. As there are many pieces of code that sets a requests.Session, we
# are creating this variable here so that a change in one affects the other. This can be removed once we merge how we do HTTP requests in
# one piece of code or once we make connection pool size configurable for each piece of code
MAX_CONNECTION_POOL_SIZE = 20



import logging
from typing import Any, List, Mapping, Optional, Tuple

from airbyte_protocol_dataclasses.models import (
    AirbyteCatalog,
    Status,
)
from fastapi import HTTPException

from airbyte_cdk.connector_builder.models import StreamRead
from airbyte_cdk.connector_builder.test_reader import TestReader
from airbyte_cdk.entrypoint import AirbyteEntrypoint
from airbyte_cdk.models import (
    AirbyteStateMessage,
    ConfiguredAirbyteCatalog,
)
from airbyte_cdk.sources.declarative.concurrent_declarative_source import (
    ConcurrentDeclarativeSource,
)
from airbyte_cdk.test.entrypoint_wrapper import AirbyteEntrypointException, EntrypointOutput


class ManifestCommandProcessor:
    _source: ConcurrentDeclarativeSource
    _logger = logging.getLogger("airbyte.manifest-server")

    def __init__(self, source: ConcurrentDeclarativeSource) -> None:
        self._source = source

    def test_read(
        self,
        config: Mapping[str, Any],
        catalog: ConfiguredAirbyteCatalog,
        state: List[AirbyteStateMessage],
        record_limit: int,
        page_limit: int,
        slice_limit: int,
    ) -> StreamRead:
        """
        Test the read method of the source.
        """
        test_read_handler = TestReader(
            max_pages_per_slice=page_limit,
            max_slices=slice_limit,
            max_record_limit=record_limit,
        )

        stream_read = test_read_handler.run_test_read(
            source=self._source,
            config=config,
            configured_catalog=catalog,
            state=state,
            stream_name=catalog.streams[0].stream.name,
            record_limit=record_limit,
        )

        return stream_read

    def check_connection(
        self,
        config: Mapping[str, Any],
    ) -> Tuple[bool, Optional[str]]:
        """
        Check the connection to the source.
        """

        spec = self._source.spec(self._logger)
        entrypoint = AirbyteEntrypoint(source=self._source)
        messages = entrypoint.check(spec, config)
        output = EntrypointOutput(
            messages=[AirbyteEntrypoint.airbyte_message_to_string(m) for m in messages],
            command=["check"],
        )
        self._raise_on_trace_message(output)

        status_messages = output.connection_status_messages
        if not status_messages or status_messages[-1].connectionStatus is None:
            return False, "Connection check did not return a status message"

        connection_status = status_messages[-1].connectionStatus
        return (
            connection_status.status == Status.SUCCEEDED,
            connection_status.message,
        )

    def discover(
        self,
        config: Mapping[str, Any],
    ) -> Optional[AirbyteCatalog]:
        """
        Discover the catalog from the source.
        """
        spec = self._source.spec(self._logger)
        entrypoint = AirbyteEntrypoint(source=self._source)
        messages = entrypoint.discover(spec, config)
        output = EntrypointOutput(
            messages=[AirbyteEntrypoint.airbyte_message_to_string(m) for m in messages],
            command=["discover"],
        )
        self._raise_on_trace_message(output)

        try:
            catalog_message = output.catalog
            return catalog_message.catalog
        except ValueError:
            # No catalog message found
            return None

    def _raise_on_trace_message(
        self,
        output: EntrypointOutput,
    ) -> None:
        """
        Raise an exception if a trace message is found.
        """
        try:
            output.raise_if_errors()
        except AirbyteEntrypointException as e:
            raise HTTPException(status_code=422, detail=e.message)



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import abc
import dataclasses
import datetime
import logging
import re
import time
from datetime import timedelta
from threading import RLock
from typing import TYPE_CHECKING, Any, Mapping, Optional
from urllib import parse

import requests
import requests_cache
from pyrate_limiter import InMemoryBucket, Limiter, RateItem, TimeClock
from pyrate_limiter import Rate as PyRateRate
from pyrate_limiter.exceptions import BucketFullException

# prevents mypy from complaining about missing session attributes in LimiterMixin
if TYPE_CHECKING:
    MIXIN_BASE = requests.Session
else:
    MIXIN_BASE = object

logger = logging.getLogger("airbyte")
logging.getLogger("pyrate_limiter").setLevel(logging.WARNING)


@dataclasses.dataclass
class Rate:
    """Call rate limit"""

    limit: int
    interval: timedelta


class CallRateLimitHit(Exception):
    def __init__(self, error: str, item: Any, weight: int, rate: str, time_to_wait: timedelta):
        """Constructor

        :param error: error message
        :param item: object passed into acquire_call
        :param weight: how many credits were requested
        :param rate: string representation of the rate violated
        :param time_to_wait: how long should wait util more call will be available
        """
        self.item = item
        self.weight = weight
        self.rate = rate
        self.time_to_wait = time_to_wait
        super().__init__(error)


class AbstractCallRatePolicy(abc.ABC):
    """Call rate policy interface.
    Should be configurable with different rules, like N per M for endpoint X. Endpoint X is matched with APIBudget.
    """

    @abc.abstractmethod
    def matches(self, request: Any) -> bool:
        """Tells if this policy matches specific request and should apply to it

        :param request:
        :return: True if policy should apply to this request, False - otherwise
        """

    @abc.abstractmethod
    def try_acquire(self, request: Any, weight: int) -> None:
        """Try to acquire request

        :param request: a request object representing a single call to API
        :param weight: number of requests to deduct from credit
        :return:
        """

    @abc.abstractmethod
    def update(
        self, available_calls: Optional[int], call_reset_ts: Optional[datetime.datetime]
    ) -> None:
        """Update call rate counting with current values

        :param available_calls:
        :param call_reset_ts:
        """


class RequestMatcher(abc.ABC):
    """Callable that help to match a request object with call rate policies."""

    @abc.abstractmethod
    def __call__(self, request: Any) -> bool:
        """

        :param request:
        :return: True if matches the provided request object, False - otherwise
        """


class HttpRequestMatcher(RequestMatcher):
    """Simple implementation of RequestMatcher for HTTP requests using HttpRequestRegexMatcher under the hood."""

    def __init__(
        self,
        method: Optional[str] = None,
        url: Optional[str] = None,
        params: Optional[Mapping[str, Any]] = None,
        headers: Optional[Mapping[str, Any]] = None,
    ):
        """Constructor

        :param method: HTTP method (e.g., "GET", "POST").
        :param url: Full URL to match.
        :param params: Dictionary of query parameters to match.
        :param headers: Dictionary of headers to match.
        """
        # Parse the URL to extract the base and path
        if url:
            parsed_url = parse.urlsplit(url)
            url_base = f"{parsed_url.scheme}://{parsed_url.netloc}"
            url_path = parsed_url.path if parsed_url.path != "/" else None
        else:
            url_base = None
            url_path = None

        # Use HttpRequestRegexMatcher under the hood
        self._regex_matcher = HttpRequestRegexMatcher(
            method=method,
            url_base=url_base,
            url_path_pattern=re.escape(url_path) if url_path else None,
            params=params,
            headers=headers,
        )

    def __call__(self, request: Any) -> bool:
        """
        :param request: A requests.Request or requests.PreparedRequest instance.
        :return: True if the request matches all provided criteria; False otherwise.
        """
        return self._regex_matcher(request)

    def __str__(self) -> str:
        return (
            f"HttpRequestMatcher(method={self._regex_matcher._method}, "
            f"url={self._regex_matcher._url_base}{self._regex_matcher._url_path_pattern.pattern if self._regex_matcher._url_path_pattern else ''}, "
            f"params={self._regex_matcher._params}, headers={self._regex_matcher._headers})"
        )


class HttpRequestRegexMatcher(RequestMatcher):
    """
    Extended RequestMatcher for HTTP requests that supports matching on:
      - HTTP method (case-insensitive)
      - URL base (scheme + netloc) optionally
      - URL path pattern (a regex applied to the path portion of the URL)
      - Query parameters (must be present)
      - Headers (header names compared case-insensitively)
    """

    def __init__(
        self,
        method: Optional[str] = None,
        url_base: Optional[str] = None,
        url_path_pattern: Optional[str] = None,
        params: Optional[Mapping[str, Any]] = None,
        headers: Optional[Mapping[str, Any]] = None,
    ):
        """
        :param method: HTTP method (e.g. "GET", "POST"); compared case-insensitively.
        :param url_base: Base URL (scheme://host) that must match.
        :param url_path_pattern: A regex pattern that will be applied to the path portion of the URL.
        :param params: Dictionary of query parameters that must be present in the request.
        :param headers: Dictionary of headers that must be present (header keys are compared case-insensitively).
        """
        self._method = method.upper() if method else None

        # Normalize the url_base if provided: remove trailing slash.
        self._url_base = url_base.rstrip("/") if url_base else None

        # Compile the URL path pattern if provided.
        self._url_path_pattern = re.compile(url_path_pattern) if url_path_pattern else None

        # Normalize query parameters to strings.
        self._params = {str(k): str(v) for k, v in (params or {}).items()}

        # Normalize header keys to lowercase.
        self._headers = {str(k).lower(): str(v) for k, v in (headers or {}).items()}

    @staticmethod
    def _match_dict(obj: Mapping[str, Any], pattern: Mapping[str, Any]) -> bool:
        """Check that every key/value in the pattern exists in the object."""
        return pattern.items() <= obj.items()

    def __call__(self, request: Any) -> bool:
        """
        :param request: A requests.Request or requests.PreparedRequest instance.
        :return: True if the request matches all provided criteria; False otherwise.
        """
        # Prepare the request (if needed) and extract the URL details.
        if isinstance(request, requests.Request):
            prepared_request = request.prepare()
        elif isinstance(request, requests.PreparedRequest):
            prepared_request = request
        else:
            return False

        # Check HTTP method.
        if self._method is not None:
            if prepared_request.method != self._method:
                return False

        # Parse the URL.
        parsed_url = parse.urlsplit(prepared_request.url)
        # Reconstruct the base: scheme://netloc
        request_url_base = f"{str(parsed_url.scheme)}://{str(parsed_url.netloc)}"
        # The path (without query parameters)
        request_path = str(parsed_url.path).rstrip("/")

        # If a base URL is provided, check that it matches.
        if self._url_base is not None:
            if request_url_base != self._url_base:
                return False

        # If a URL path pattern is provided, ensure the path matches the regex.
        if self._url_path_pattern is not None:
            if not self._url_path_pattern.search(request_path):
                return False

        # Check query parameters.
        if self._params:
            query_params = dict(parse.parse_qsl(str(parsed_url.query)))
            if not self._match_dict(query_params, self._params):
                return False

        # Check headers (normalize keys to lower-case).
        if self._headers:
            req_headers = {k.lower(): v for k, v in prepared_request.headers.items()}
            if not self._match_dict(req_headers, self._headers):
                return False

        return True

    def __str__(self) -> str:
        regex = self._url_path_pattern.pattern if self._url_path_pattern else None
        return (
            f"HttpRequestRegexMatcher(method={self._method}, url_base={self._url_base}, "
            f"url_path_pattern={regex}, params={self._params}, headers={self._headers})"
        )


class BaseCallRatePolicy(AbstractCallRatePolicy, abc.ABC):
    def __init__(self, matchers: list[RequestMatcher]):
        self._matchers = matchers

    def matches(self, request: Any) -> bool:
        """Tell if this policy matches specific request and should apply to it

        :param request:
        :return: True if policy should apply to this request, False - otherwise
        """

        if not self._matchers:
            return True
        return any(matcher(request) for matcher in self._matchers)


class UnlimitedCallRatePolicy(BaseCallRatePolicy):
    """
    This policy is for explicit unlimited call rates.
    It can be used when we want to match a specific group of requests and don't apply any limits.

    Example:

    APICallBudget(
        [
            UnlimitedCallRatePolicy(
                matchers=[HttpRequestMatcher(url="/some/method", headers={"sandbox": true})],
            ),
            FixedWindowCallRatePolicy(
                matchers=[HttpRequestMatcher(url="/some/method")],
                next_reset_ts=datetime.now(),
                period=timedelta(hours=1)
                call_limit=1000,
            ),
        ]
    )

    The code above will limit all calls to /some/method except calls that have header sandbox=True
    """

    def try_acquire(self, request: Any, weight: int) -> None:
        """Do nothing"""

    def update(
        self, available_calls: Optional[int], call_reset_ts: Optional[datetime.datetime]
    ) -> None:
        """Do nothing"""


class FixedWindowCallRatePolicy(BaseCallRatePolicy):
    def __init__(
        self,
        next_reset_ts: datetime.datetime,
        period: timedelta,
        call_limit: int,
        matchers: list[RequestMatcher],
    ):
        """A policy that allows {call_limit} calls within a {period} time interval

        :param next_reset_ts: next call rate reset time point
        :param period: call rate reset period
        :param call_limit:
        :param matchers:
        """

        self._next_reset_ts = next_reset_ts
        self._offset = period
        self._call_limit = call_limit
        self._calls_num = 0
        self._lock = RLock()
        super().__init__(matchers=matchers)

    def try_acquire(self, request: Any, weight: int) -> None:
        if weight > self._call_limit:
            raise ValueError("Weight can not exceed the call limit")
        if not self.matches(request):
            raise ValueError("Request does not match the policy")

        with self._lock:
            self._update_current_window()

            if self._calls_num + weight > self._call_limit:
                reset_in = self._next_reset_ts - datetime.datetime.now()
                error_message = (
                    f"reached maximum number of allowed calls {self._call_limit} "
                    f"per {self._offset} interval, next reset in {reset_in}."
                )
                raise CallRateLimitHit(
                    error=error_message,
                    item=request,
                    weight=weight,
                    rate=f"{self._call_limit} per {self._offset}",
                    time_to_wait=reset_in,
                )

            self._calls_num += weight

    def __str__(self) -> str:
        matcher_str = ", ".join(f"{matcher}" for matcher in self._matchers)
        return (
            f"FixedWindowCallRatePolicy(call_limit={self._call_limit}, period={self._offset}, "
            f"calls_used={self._calls_num}, next_reset={self._next_reset_ts}, "
            f"matchers=[{matcher_str}])"
        )

    def update(
        self, available_calls: Optional[int], call_reset_ts: Optional[datetime.datetime]
    ) -> None:
        """Update call rate counters, by default, only reacts to decreasing updates of available_calls and changes to call_reset_ts.
        We ignore updates with available_calls > current_available_calls to support call rate limits that are lower than API limits.

        :param available_calls:
        :param call_reset_ts:
        """
        with self._lock:
            self._update_current_window()
            current_available_calls = self._call_limit - self._calls_num

            if available_calls is not None and current_available_calls > available_calls:
                logger.debug(
                    "got rate limit update from api, adjusting available calls from %s to %s",
                    current_available_calls,
                    available_calls,
                )
                self._calls_num = self._call_limit - available_calls

            if call_reset_ts is not None and call_reset_ts != self._next_reset_ts:
                logger.debug(
                    "got rate limit update from api, adjusting reset time from %s to %s",
                    self._next_reset_ts,
                    call_reset_ts,
                )
                self._next_reset_ts = call_reset_ts

    def _update_current_window(self) -> None:
        now = datetime.datetime.now()
        if now > self._next_reset_ts:
            logger.debug("started new window, %s calls available now", self._call_limit)
            self._next_reset_ts = self._next_reset_ts + self._offset
            self._calls_num = 0


class MovingWindowCallRatePolicy(BaseCallRatePolicy):
    """
    Policy to control requests rate implemented on top of PyRateLimiter lib.
    The main difference between this policy and FixedWindowCallRatePolicy is that the rate-limiting window
    is moving along requests that we made, and there is no moment when we reset an available number of calls.
    This strategy requires saving of timestamps of all requests within a window.
    """

    def __init__(self, rates: list[Rate], matchers: list[RequestMatcher]):
        """Constructor

        :param rates: list of rates, the order is important and must be ascending
        :param matchers:
        """
        if not rates:
            raise ValueError("The list of rates can not be empty")
        pyrate_rates = [
            PyRateRate(limit=rate.limit, interval=int(rate.interval.total_seconds() * 1000))
            for rate in rates
        ]
        self._bucket = InMemoryBucket(pyrate_rates)
        # Limiter will create the background task that clears old requests in the bucket
        self._limiter = Limiter(self._bucket)
        super().__init__(matchers=matchers)

    def try_acquire(self, request: Any, weight: int) -> None:
        if not self.matches(request):
            raise ValueError("Request does not match the policy")

        try:
            self._limiter.try_acquire(request, weight=weight)
        except BucketFullException as exc:
            item = self._limiter.bucket_factory.wrap_item(request, weight)
            assert isinstance(item, RateItem)

            with self._limiter.lock:
                time_to_wait = self._bucket.waiting(item)
                assert isinstance(time_to_wait, int)

                raise CallRateLimitHit(
                    error=str(exc.meta_info["error"]),
                    item=request,
                    weight=int(exc.meta_info["weight"]),
                    rate=str(exc.meta_info["rate"]),
                    time_to_wait=timedelta(milliseconds=time_to_wait),
                )

    def update(
        self, available_calls: Optional[int], call_reset_ts: Optional[datetime.datetime]
    ) -> None:
        """Adjust call bucket to reflect the state of the API server

        :param available_calls:
        :param call_reset_ts:
        :return:
        """
        if (
            available_calls is not None and call_reset_ts is None
        ):  # we do our best to sync buckets with API
            if available_calls == 0:
                with self._limiter.lock:
                    items_to_add = self._bucket.count() < self._bucket.rates[0].limit
                    if items_to_add > 0:
                        now: int = TimeClock().now()  # type: ignore[no-untyped-call]
                        self._bucket.put(RateItem(name="dummy", timestamp=now, weight=items_to_add))
        # TODO: add support if needed, it might be that it is not possible to make a good solution for this case
        # if available_calls is not None and call_reset_ts is not None:
        #     ts = call_reset_ts.timestamp()

    def __str__(self) -> str:
        """Return a human-friendly description of the moving window rate policy for logging purposes."""
        rates_info = ", ".join(
            f"{rate.limit} per {timedelta(milliseconds=rate.interval)}"
            for rate in self._bucket.rates
        )
        current_bucket_count = self._bucket.count()
        matcher_str = ", ".join(f"{matcher}" for matcher in self._matchers)
        return (
            f"MovingWindowCallRatePolicy(rates=[{rates_info}], current_bucket_count={current_bucket_count}, "
            f"matchers=[{matcher_str}])"
        )


class AbstractAPIBudget(abc.ABC):
    """Interface to some API where a client allowed to have N calls per T interval.

    Important: APIBudget is not doing any API calls, the end user code is responsible to call this interface
        to respect call rate limitation of the API.

    It supports multiple policies applied to different group of requests. To distinct these groups we use RequestMatchers.
    Individual policy represented by MovingWindowCallRatePolicy and currently supports only moving window strategy.
    """

    @abc.abstractmethod
    def acquire_call(
        self, request: Any, block: bool = True, timeout: Optional[float] = None
    ) -> None:
        """Try to get a call from budget, will block by default

        :param request:
        :param block: when true (default) will block the current thread until call credit is available
        :param timeout: if set will limit maximum time in block, otherwise will wait until credit is available
        :raises: CallRateLimitHit - when no credits left and if timeout was set the waiting time exceed the timeout
        """

    @abc.abstractmethod
    def get_matching_policy(self, request: Any) -> Optional[AbstractCallRatePolicy]:
        """Find matching call rate policy for specific request"""

    @abc.abstractmethod
    def update_from_response(self, request: Any, response: Any) -> None:
        """Update budget information based on response from API

        :param request: the initial request that triggered this response
        :param response: response from the API
        """


class APIBudget(AbstractAPIBudget):
    """Default APIBudget implementation"""

    def __init__(
        self, policies: list[AbstractCallRatePolicy], maximum_attempts_to_acquire: int = 100000
    ) -> None:
        """Constructor

        :param policies: list of policies in this budget
        :param maximum_attempts_to_acquire: number of attempts before throwing hit ratelimit exception, we put some big number here
         to avoid situations when many threads compete with each other for a few lots over a significant amount of time
        """

        self._policies = policies
        self._maximum_attempts_to_acquire = maximum_attempts_to_acquire

    def _extract_endpoint(self, request: Any) -> str:
        """Extract the endpoint URL from the request if available."""
        endpoint = None
        try:
            # If the request is already a PreparedRequest, it should have a URL.
            if isinstance(request, requests.PreparedRequest):
                endpoint = request.url
            # If it's a requests.Request, we call prepare() to extract the URL.
            elif isinstance(request, requests.Request):
                prepared = request.prepare()
                endpoint = prepared.url
        except Exception as e:
            logger.debug(f"Error extracting endpoint: {e}")
        if endpoint:
            return endpoint
        return "unknown endpoint"

    def get_matching_policy(self, request: Any) -> Optional[AbstractCallRatePolicy]:
        for policy in self._policies:
            if policy.matches(request):
                return policy
        return None

    def acquire_call(
        self, request: Any, block: bool = True, timeout: Optional[float] = None
    ) -> None:
        """Try to get a call from budget, will block by default.
        Matchers will be called sequentially in the same order they were added.
        The first matcher that returns True will

        :param request: the API request
        :param block: when True (default) will block until a call credit is available
        :param timeout: if provided, limits maximum waiting time; otherwise, waits indefinitely
        :raises: CallRateLimitHit if the call credit cannot be acquired within the timeout
        """

        policy = self.get_matching_policy(request)
        endpoint = self._extract_endpoint(request)
        if policy:
            logger.debug(f"Acquiring call for endpoint {endpoint} using policy: {policy}")
            self._do_acquire(request=request, policy=policy, block=block, timeout=timeout)
        elif self._policies:
            logger.debug(
                f"No policies matched for endpoint {endpoint} (request: {request}). Allowing call by default."
            )

    def update_from_response(self, request: Any, response: Any) -> None:
        """Update budget information based on the API response.

        :param request: the initial request that triggered this response
        :param response: response from the API
        """
        pass

    def _do_acquire(
        self, request: Any, policy: AbstractCallRatePolicy, block: bool, timeout: Optional[float]
    ) -> None:
        """Internal method to try to acquire a call credit.

        :param request: the API request
        :param policy: the matching rate-limiting policy
        :param block: indicates whether to block until a call credit is available
        :param timeout: maximum time to wait if blocking
        :raises: CallRateLimitHit if unable to acquire a call credit
        """
        last_exception = None
        endpoint = self._extract_endpoint(request)
        # sometimes we spend all budget before a second attempt, so we have a few more attempts
        for attempt in range(1, self._maximum_attempts_to_acquire):
            try:
                policy.try_acquire(request, weight=1)
                return
            except CallRateLimitHit as exc:
                last_exception = exc
                if block:
                    if timeout is not None:
                        time_to_wait = min(timedelta(seconds=timeout), exc.time_to_wait)
                    else:
                        time_to_wait = exc.time_to_wait
                    # Ensure we never sleep for a negative duration.
                    time_to_wait = max(timedelta(0), time_to_wait)
                    logger.debug(
                        f"Policy {policy} reached call limit for endpoint {endpoint} ({exc.rate}). "
                        f"Sleeping for {time_to_wait} on attempt {attempt}."
                    )
                    time.sleep(time_to_wait.total_seconds())
                else:
                    logger.debug(
                        f"Policy {policy} reached call limit for endpoint {endpoint} ({exc.rate}) "
                        f"and blocking is disabled."
                    )
                    raise

        if last_exception:
            logger.debug(
                f"Exhausted all {self._maximum_attempts_to_acquire} attempts to acquire a call for endpoint {endpoint} "
                f"using policy: {policy}"
            )
            raise last_exception


class HttpAPIBudget(APIBudget):
    """Implementation of AbstractAPIBudget for HTTP"""

    def __init__(
        self,
        ratelimit_reset_header: str = "ratelimit-reset",
        ratelimit_remaining_header: str = "ratelimit-remaining",
        status_codes_for_ratelimit_hit: list[int] = [429],
        **kwargs: Any,
    ):
        """Constructor

        :param ratelimit_reset_header: name of the header that has a timestamp of the next reset of call budget
        :param ratelimit_remaining_header: name of the header that has the number of calls left
        :param status_codes_for_ratelimit_hit: list of HTTP status codes that signal about rate limit being hit
        """
        self._ratelimit_reset_header = ratelimit_reset_header
        self._ratelimit_remaining_header = ratelimit_remaining_header
        self._status_codes_for_ratelimit_hit = status_codes_for_ratelimit_hit
        super().__init__(**kwargs)

    def update_from_response(self, request: Any, response: Any) -> None:
        policy = self.get_matching_policy(request)
        if not policy:
            return

        if isinstance(response, requests.Response):
            available_calls = self.get_calls_left_from_response(response)
            reset_ts = self.get_reset_ts_from_response(response)
            policy.update(available_calls=available_calls, call_reset_ts=reset_ts)

    def get_reset_ts_from_response(
        self, response: requests.Response
    ) -> Optional[datetime.datetime]:
        if response.headers.get(self._ratelimit_reset_header):
            return datetime.datetime.fromtimestamp(
                int(response.headers[self._ratelimit_reset_header])
            )
        return None

    def get_calls_left_from_response(self, response: requests.Response) -> Optional[int]:
        if response.headers.get(self._ratelimit_remaining_header):
            return int(response.headers[self._ratelimit_remaining_header])

        if response.status_code in self._status_codes_for_ratelimit_hit:
            return 0

        return None


class LimiterMixin(MIXIN_BASE):
    """Mixin class that adds rate-limiting behavior to requests."""

    def __init__(
        self,
        api_budget: AbstractAPIBudget,
        **kwargs: Any,
    ):
        self._api_budget = api_budget
        super().__init__(**kwargs)  # type: ignore # Base Session doesn't take any kwargs

    def send(self, request: requests.PreparedRequest, **kwargs: Any) -> requests.Response:
        """Send a request with rate-limiting."""
        self._api_budget.acquire_call(request)
        response = super().send(request, **kwargs)
        self._api_budget.update_from_response(request, response)
        return response


class LimiterSession(LimiterMixin, requests.Session):
    """Session that adds rate-limiting behavior to requests."""


class CachedLimiterSession(requests_cache.CacheMixin, LimiterMixin, requests.Session):
    """Session class with caching and rate-limiting behavior."""



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import re
from typing import Any, Dict, Mapping, Set, Tuple, Union

from airbyte_cdk.sources.declarative.parsers.custom_exceptions import (
    CircularReferenceException,
    UndefinedReferenceException,
)

REF_TAG = "$ref"


class ManifestReferenceResolver:
    """
    An incoming manifest can contain references to values previously defined.
    This parser will dereference these values to produce a complete ConnectionDefinition.

    References can be defined using a #/<arg> string.
    ```
    key: 1234
    reference: "#/key"
    ```
    will produce the following definition:
    ```
    key: 1234
    reference: 1234
    ```
    This also works with objects:
    ```
    key_value_pairs:
      k1: v1
      k2: v2
    same_key_value_pairs: "#/key_value_pairs"
    ```
    will produce the following definition:
    ```
    key_value_pairs:
      k1: v1
      k2: v2
    same_key_value_pairs:
      k1: v1
      k2: v2
    ```

    The $ref keyword can be used to refer to an object and enhance it with addition key-value pairs
    ```
    key_value_pairs:
      k1: v1
      k2: v2
    same_key_value_pairs:
      $ref: "#/key_value_pairs"
      k3: v3
    ```
    will produce the following definition:
    ```
    key_value_pairs:
      k1: v1
      k2: v2
    same_key_value_pairs:
      k1: v1
      k2: v2
      k3: v3
    ```

    References can also point to nested values.
    Nested references are ambiguous because one could define a key containing with `.`
    in this example, we want to refer to the limit key in the dict object:
    ```
    dict:
        limit: 50
    limit_ref: "#/dict/limit"
    ```
    will produce the following definition:
    ```
    dict
        limit: 50
    limit-ref: 50
    ```

    whereas here we want to access the `nested/path` value.
    ```
    nested:
        path: "first one"
    nested/path: "uh oh"
    value: "#/nested/path
    ```
    will produce the following definition:
    ```
    nested:
        path: "first one"
    nested/path: "uh oh"
    value: "uh oh"
    ```

    to resolve the ambiguity, we try looking for the reference key at the top level, and then traverse the structs downward
    until we find a key with the given path, or until there is nothing to traverse.
    """

    def preprocess_manifest(self, manifest: Mapping[str, Any]) -> Dict[str, Any]:
        """
        :param manifest: incoming manifest that could have references to previously defined components
        :return:
        """
        return self._evaluate_node(manifest, manifest, set())  # type: ignore[no-any-return]

    def _evaluate_node(self, node: Any, manifest: Mapping[str, Any], visited: Set[Any]) -> Any:
        if isinstance(node, dict):
            evaluated_dict = {
                k: self._evaluate_node(v, manifest, visited)
                for k, v in node.items()
                if not self._is_ref_key(k)
            }
            if REF_TAG in node:
                # The node includes a $ref key, so we splat the referenced value(s) into the evaluated dict
                evaluated_ref = self._evaluate_node(node[REF_TAG], manifest, visited)
                if not isinstance(evaluated_ref, dict):
                    return evaluated_ref
                else:
                    # The values defined on the component take precedence over the reference values
                    return evaluated_ref | evaluated_dict
            else:
                return evaluated_dict
        elif isinstance(node, list):
            return [self._evaluate_node(v, manifest, visited) for v in node]
        elif self._is_ref(node):
            if node in visited:
                raise CircularReferenceException(node)
            visited.add(node)
            ret = self._evaluate_node(self._lookup_ref_value(node, manifest), manifest, visited)
            visited.remove(node)
            return ret
        else:
            return node

    def _lookup_ref_value(self, ref: str, manifest: Mapping[str, Any]) -> Any:
        ref_match = re.match(r"#/(.*)", ref)
        if not ref_match:
            raise ValueError(f"Invalid reference format {ref}")
        try:
            path = ref_match.groups()[0]
            return self._read_ref_value(path, manifest)
        except (AttributeError, KeyError, IndexError):
            raise UndefinedReferenceException(path, ref)

    @staticmethod
    def _is_ref(node: Any) -> bool:
        return isinstance(node, str) and node.startswith("#/")

    @staticmethod
    def _is_ref_key(key: str) -> bool:
        return bool(key == REF_TAG)

    @staticmethod
    def _read_ref_value(ref: str, manifest_node: Mapping[str, Any]) -> Any:
        """
        Read the value at the referenced location of the manifest.

        References are ambiguous because one could define a key containing `/`
        In this example, we want to refer to the `limit` key in the `dict` object:
            dict:
                limit: 50
            limit_ref: "#/dict/limit"

        Whereas here we want to access the `nested/path` value.
          nested:
            path: "first one"
          nested/path: "uh oh"
          value: "#/nested/path"

        To resolve the ambiguity, we try looking for the reference key at the top level, and then traverse the structs downward
        until we find a key with the given path, or until there is nothing to traverse.

        Consider the path foo/bar/baz. To resolve the ambiguity, we first try 'foo/bar/baz' in its entirety as a top-level key. If this
        fails, we try 'foo' as the top-level key, and if this succeeds, pass 'bar/baz' on as the key to be tried at the next level.
        """
        while ref:
            try:
                return manifest_node[ref]
            except (KeyError, TypeError):
                head, ref = _parse_path(ref)
                manifest_node = manifest_node[head]  # type: ignore # Couldn't figure out how to fix this since manifest_node can get reassigned into other types like lists
        return manifest_node


def _parse_path(ref: str) -> Tuple[Union[str, int], str]:
    """
    Return the next path component, together with the rest of the path.

    A path component may be a string key, or an int index.

    >>> _parse_path("foo/bar")
    "foo", "bar"
    >>> _parse_path("foo/7/8/bar")
    "foo", "7/8/bar"
    >>> _parse_path("7/8/bar")
    7, "8/bar"
    >>> _parse_path("8/bar")
    8, "bar"
    >>> _parse_path("8foo/bar")
    "8foo", "bar"
    """
    match = re.match(r"([^/]*)/?(.*)", ref)
    if match:
        first, rest = match.groups()
        try:
            return int(first), rest
        except ValueError:
            return first, rest
    else:
        raise ValueError(f"Invalid path {ref} specified")
